//WBasicShape.java
package com.tanksoftware.graphics;

import java.awt.Graphics;
import java.awt.geom.Point2D;
import java.awt.Color;

/**
 * The standard shape that all shapes in the tanksoftware.shape package extend.
 * It provides the basic framework for doing other, more specialised operations.
 *
 * <p>Title: WBasicShape</p>
 * <p>Description: This class provides the basic framework for a game or application
 * which uses shapes.  It, and the other classes in the tanksoftware.geom package
 * provide a simplified way of drawing and manipulating basic shapes such as
 * Circles, and Polygons.  It sets out some basic values and methods, while the
 * actuall drawing of the shape is left to that shapes specific implementation.</p>
 * <p>An example of this class in use can be found in the Tank Software program Tank Commander 2</p>
 * <p>Copyright: Copyright (c) 2002, free non-commercial use is permitted.</p>
 * <p>Company: Tank Software</p>
 * @author William Denniss
 * @version 1.0
 */

public abstract class WBasicShape {


	public double XCenter;
	public double YCenter;
	public double angle = 0;


    /**
     * Moves this shape to a new absolute location.
     *
     * @param newXCenter The absolute value that will become the new X coordinate
     * of the centerfor the shape.
     * @param newYCenter The absolute value that will become the new Y coordinate
     * of the centerfor the shape.
     */
	public void move (double newXCenter, double newYCenter) {
		XCenter = newXCenter;
		YCenter = newYCenter;
	}

    /**
     * Moves this shape to a new location as specified by the given relitive
     * coordinates.
     *
     * @param XOffset The value by which the X coordinate of the shape will be
     * offset (can be negative).
     * @param YOffset The value by which the Y coordinate of the shape will be
     * offset (can be negative).
     */
	public void moveOffset (double XOffset, double YOffset) {
		XCenter = XCenter + XOffset;
		YCenter = YCenter + YOffset;
	}

    /**
     * Moves this shape by the given magnitude in the direction that the shape
     * is currently pointing.
     *
     * @param magnitude The distance the shape will be moved.
     */
	public void moveOffset (double magnitude) {
		moveOffset(magnitude * Math.sin(Math.toRadians(angle)), magnitude * Math.cos(Math.toRadians(angle)));
	}

    /**
     * Rotates the shape by the given amount
     *
     * @param degrees The amount this shape is rotated.
     */
	public void rotate (double degrees) {
		angle = angle + degrees;
	}

    /**
     * Calculates the horisontal and vertical components of an increase of a
     * magnitude by an angle.
     *
     * @param magnitude The distance moved in the given direction.
     * @param angle The direction the point is moved
     * @return A Point2D.Double object which contains as it's X parameter the
     * distance moved horisontally, and in it's Y perameter the distance moved
     * vertically.
     */
	public static Point2D.Double calcOffset (double magnitude, double angle) {
		double x = magnitude * Math.sin(Math.toRadians(angle));
		double y = magnitude * Math.cos(Math.toRadians(angle));
		return new Point2D.Double (x,y);

	}

    /**
     * Draws the shape filled.
     *
     * @param g The Graphics class to which the shape will be drawn to.
     */
   	public abstract void drawFilled(Graphics g);

    /**
     * Removes the filled shape and replaces it with a given colour.
     *
     * @param g The Graphics class to which the overlaying form of the shape
     * will be drawn to.
     * @param colour The background colour that will replace the shape.
     */
	public abstract void removeFilled(Graphics g, Color colour);

    /**
     * Enlarges thise shape by a given factor.
     *
     * @param factor The factor by which this shape will be enlarged.
     */
	public abstract void enlarge (double factor);


    /**
     * Gets the current angle at which this shape is pointing
     *
     * @return The current angle.
     */
	public double getAngle () {
		return angle;
	}


}