//WPCircle.java
package com.tanksoftware.graphics;

import java.awt.Graphics;
import java.awt.Color;
import java.awt.geom.Point2D;


/**
 * An implementation of the WShape class that is a circle.
 *
 * <p>Title: WShape</p>
 * <p>Description: This is a circle which has a defined center point, and a
 * rotation point by which the center point can be rotated.
 * <p>An example of this class in use can be found in the Tank Software program Tank Commander 2</p>
 * <p>Copyright: Copyright (c) 2002, free non-commercial use is permitted.</p>
 * <p>Company: Tank Software</p>
 * @author William Denniss
 * @version 1.0
 */
public class WCircle extends WShape{


	public double diameter;
	public double XOffset;
	public double YOffset;

    /**
     * Creates a new WCircle object with a center point, radius of the circle, and
     * rotation point (the point by which the center point will be rotated in the
     * <code>rotate</code> method.
     *
     * @param x The x-coordinate of the center point.
     * @param y The y-coordinate of the center point.
     * @param diameter The diameter.
     * @param XOffset The x-coordinate of the rotation point.
     * @param YOffset The y-coordinate of the rotation point.
     */
	public WCircle (double x, double y, double diameter, double XOffset, double YOffset) {
		super (x, y);
		this.diameter = diameter;
		this.XOffset = XOffset;
		this.YOffset = YOffset;
	}


    /**
     * Increase the size of this polygon by the given factor.
     *
     * @param factor The factor by which to enlarge this polygon.
     */
	public void enlarge (double factor) {
		XOffset = XOffset * factor;
		YOffset = YOffset * factor;
		diameter = diameter * factor;
	}




    /**
     * Rotates the center of this WCircle around it's rotation point.
     *
     * @param degrees The amount this shape is rotated in.
     * @see <code>WShape.rotatePoint (double x, double y, double degrees) </code>for how the
     * points are actually rotated.
     */
	public void rotate (double degrees){
		super.rotate(degrees);
		Point2D.Double offset = WShape.rotatePoint(XOffset, YOffset, degrees);
		XOffset = offset.x;
		YOffset = offset.y;
	}

    /**
     * Draws this circle as a hollow shape.
     *
     * @param g The Graphics object to which it will be drawn.
     */
	public void drawHollow(Graphics g){
		g.drawOval((int) XCenter,(int) YCenter,(int) diameter,(int) diameter);
	}

    /**
     * Draws a silhouette of this circle as a hollow shape in a given colour
     *
     * @param g The Graphics object to which it will be drawn.
     * @param removeColour The colour of the silhouette.  If this is the background
     * colour, it will effectivly "remove" this shape.
     */
	public void removeHollow(Graphics g, Color removeColour){
		Color oldColour = g.getColor();
		g.setColor(removeColour);
		g.drawOval((int) XCenter,(int) YCenter,(int) diameter,(int) diameter);
		g.setColor(oldColour);
	}

    /**
     * Draws this circle as a solid shape.
     *
     * @param g The Graphics object to which it will be drawn.
     */
	public void drawFilled(Graphics g){
		if (colourSet) g.setColor(colour);
		g.fillOval((int) (XCenter + XOffset - diameter/2),(int) (YCenter+YOffset- diameter/2),(int) diameter, (int) diameter);

	}

    /**
     * Draws a silhouette of this circle as a solid shape in a given colour
     *
     * @param g The Graphics object to which it will be drawn.
     * @param removeColour The colour of the silhouette.  If this is the background
     * colour, it will effectivly "remove" this shape.
     */
	public void removeFilled(Graphics g, Color removeColour){
		Color oldColour = g.getColor();
		g.setColor(removeColour);
		g.fillOval((int) XCenter,(int) YCenter,(int) diameter,(int) diameter);
		g.setColor(oldColour);

	}

    /**
     * Returns the String representation of this WCircle.
     *
     * @return the String representation of this WCircle.
     */
	public String toString () {
		return "WCircle  " + "\n" + "XCenter: " + XCenter + ", YCenter: " + YCenter + ", Radius " + diameter + ", XOffset: " + XOffset + ", YOffset: " + YOffset;
	}


}