//WPoly.java
package com.tanksoftware.graphics;

import java.awt.Polygon;
import java.awt.Graphics;
import java.awt.Color;
import java.awt.Rectangle;
import java.awt.geom.Point2D;

/**
 * An implementation of the WShape class that is a polygon.
 *
 * <p>Title: WPoly</p>
 * <p>Description: This is a polygon which has a defined center
 * The X and Y Coordinates are stored as offsets to this center as opposed to the
 * java.awt.Polygon class which stores them as absolute co-ordinates.
 * This allows the polygon to be rotated around the center
 * aswell as being able to be moved just by moving the center </p>
 * <p>An example of this class in use can be found in the Tank Software program Tank Commander 2</p>
 * <p>Copyright: Copyright (c) 2002, free non-commercial use is permitted.</p>
 * <p>Company: Tank Software</p>
 * @author William Denniss
 * @version 1.0
 */


public class WPoly extends WShape {

	public double [] XCoordinatesOffset;
	public double [] YCoordinatesOffset;
	public int nPoints;


    /**
     * Creates a new WPoly using the data from a java.awt.Polygon, and the given
     * center coordinate.
     *
     * @param poly The polygon with absoulate positioning data that will be used
     * to create this WPoly (with it's positioning data relitive to the center
     * coordinate.)
     * @param XCenter The X-coordinate of the center point.
     * @param YCenter The Y-coordinate of the center point.
     */
	public WPoly (Polygon poly, double XCenter, double YCenter) {
		super (XCenter, YCenter);

		nPoints = poly.npoints;

		XCoordinatesOffset = new double [nPoints];
		YCoordinatesOffset = new double [nPoints];
		for (int i = 0; i < nPoints; i++) {
			XCoordinatesOffset[i] = poly.xpoints[i] - XCenter;
			YCoordinatesOffset[i] = poly.ypoints[i] - YCenter;
		}

	}

    /**
     * Creates a WPoly where the data for the X and Y coordinates, is already
     * relitive to the Center Point.
     *
     * @param XCoordinatesOffset The array containing the X coordinates that are
     * RELITIVE to the center coordinate.  This is not the same X Coordinate
     * that is used in the java.awt.Polygon class.  To use that, see the
     * WPoly (Polygon poly, double XCenter, double YCenter) constructor.
     * @param YCoordinatesOffset the array containing the Y coordinates that are
     * relitive to the center coordinate, like the XCoordinatesOffset parameter.
     * @param nPoints The number of points in this polygon
     * @param XCenter The X-coordinate of the center point.
     * @param YCenter The Y-coordinate of the center point.
     */
	public WPoly (double [] XCoordinatesOffset, double [] YCoordinatesOffset, int nPoints, double XCenter, double YCenter) {
		super (XCenter, YCenter);

		this.XCoordinatesOffset = XCoordinatesOffset;
		this.YCoordinatesOffset = YCoordinatesOffset;
		this.nPoints = nPoints;

	}

    /**
     * Increase the size of this polygon by the given factor.
     *
     * @param factor The factor by which to enlarge this polygon.
     */
	public void enlarge (double factor) {
		for (int i = 0; i<XCoordinatesOffset.length;i++) {
			XCoordinatesOffset[i] = XCoordinatesOffset[i] * factor;
			YCoordinatesOffset[i] = YCoordinatesOffset[i] * factor;
		}
	}


    /**
     * Checks to see if the passed java.awt.Polygon intercects this polygon
     * using the fastIntercect implementation.
     *
     * @param poly The polygon that will be checked with this one for an intercect.
     * @return true if they do intercect, false if they don't.
     * @see <code> fastIntercect (Polygon poly1, Polygon poly2) </code> for more info on the
     * fastIntercect implementation
     */
	public boolean fastIntercect (Polygon poly) {
		return fastIntercect(toPolygon(), poly);

	}

    /**
     * Checks to see if the two given polygons intercect.  This is done by checking
     * to see if any of the points of the first polygon lie in the second, and
     * vice versa.  It is a fast operation, as there are usually not too many points
     * to a polygon, however it is not 100% accurate, as it is possible for two
     * polygons to intercect yet have no points that reside in each other.  If
     * this method returns true, then it is 100% certain that the 2 polygons
     * intercect but if it returns false, there is an assumption that they do.
     * For this reason, use this method of determining intercection only if
     * time is of importance.
     *
     * @param poly1 The first of two polygons that will be checked wether or not
     * they intercect.
     * @param poly2 The second of the two polygons.
     * @return true if they do intercect, false if they don't.
     * @see <code>accurateIntercect (Polygon poly1, Polygon poly2, int XStep, int YStep)
     * </code> for a more accurate but slower intercetion method.
     */
	public static boolean fastIntercect (Polygon poly1, Polygon poly2) {

		for (int i = 0; i<poly1.npoints; i++) {
			if (poly2.contains(poly1.xpoints[i], poly1.ypoints[i]))
				return true;
		}
		for (int i = 0; i<poly2.npoints; i++) {
			if (poly1.contains(poly2.xpoints[i], poly2.ypoints[i]))
				return true;
		}

		return false;
	}

    /**
     * Checks to see if the passed java.awt.Polygon intercects this WPoly
     * using the accurate implementation.  This method is completally accurate.
     * If it returns true, then the chances are almost 1, that they do intercect.
     *
     * @param poly
     * @return true if they do intercect, false if they don't.
     * @see <code>accurateIntercect (Polygon poly1, Polygon poly2, int XStep, int YStep)</code>
     * for info on how it does this.
     */
	public boolean accurateIntercect (Polygon poly) {
		return accurateIntercect(poly,1,1);
	}

    /**
     * Checks to see if the passed java.awt.Polygon intercects this WPoly
     * using the accurate implementation.  By default, the step rates are 1, this
     * garentees that every coordinate of the polygon will be checked, but they
     * can be changed to 2, where every second coordinate would be checked. The
     * accuracy of this method is 100% is the steps are both set to 1, but it is
     * a varying degree less when they are not set to 1.
     *
     * @param poly The Polygon that will be checked against this WPoly for intercection.
     * @param XStep The step rate of the X coordinates that will be checked.  ie.
     * if set to 1, every coordinate will be checked, if set to 2, every second etc.
     * @param YStep The step rate of the Y coordinates that will be checked.  ie.
     * if set to 1, every coordinate will be checked, if set to 2, every second etc.
     * @return true if they do intercect, false if they don't.
     * @see <code>accurateIntercect (Polygon poly1, Polygon poly2, int XStep, int YStep)</code>
     * for info on how it does this.
     */
	public boolean accurateIntercect (Polygon poly, int XStep, int YStep) {
			return accurateIntercect(toPolygon(), poly, XStep, YStep);
	}

    /**
     * Checks to see if the first passed java.awt.Polygon intercects the second
     * passed java.awt.Polygon using the accurate implementation.  This method is completally accurate.
     * If it returns true, then the chances are almost 1, that they do intercect.
     *
     * @param poly1 The first of 2 polygons to be checked for intercection.
     * @param poly2 The second of the 2 polygons to be checked for intercection.
     * @return true if they do intercect, false if they don't.
     * @see <code>accurateIntercect (Polygon poly1, Polygon poly2, int XStep, int YStep)</code>
     * for info on how it does this.
     */
	public static boolean accurateIntercect (Polygon poly1, Polygon poly2) {
        //Check every coordinate withing the bounds of this polygon, return true if a coordinate is contained in both this polygon and the parsed one
		return accurateIntercect(poly1, poly2, 1, 1);
	}

    /**
     * This is a more accurate way of determining if two polygons interect.  Unlike the
     * fastIntercect method, this one checks every point in the bounding box of the first
     * polygon to see if it is in both first and second polygons (or every nth if
     * specified in the XStep and YStep parameters.  If the X and Y steps are 1, then
     * this method is extreemly accurate, as that means that every single coordinate that
     * resites in the containing square (bounding box) of poly1 is checked for intercection.
     * The disadvantage of this is that if it is a large polygon, it could take a while
     * on slower computers.  If speed is a concern, but the fastIntercect method doesn't
     * meet your needs, then you can consider varying the X and Y step rates.  The
     * larger the step rates though, the larger the 2 polygons will need to be to maintin
     * accuracy of this method.  Change at own discresion.
     *
     * @param poly1 The first of 2 polygons to be checked for intercection.
     * @param poly2 The second of the 2 polygons to be checked for intercection.
     * @param XStep The step rate of the X coordinates that will be checked.  ie.
     * if set to 1, every coordinate will be checked, if set to 2, every second etc.
     * @param YStep The step rate of the Y coordinates that will be checked.  ie.
     * if set to 1, every coordinate will be checked, if set to 2, every second etc.
     * @return true if they do intercect, false if they don't.
     * @see <code>fastIntercect (Polygon poly1, Polygon poly2) </code>for a less CPU intensive
     * implementation of the intercect method.
     */
	public static boolean accurateIntercect (Polygon poly1, Polygon poly2, int XStep, int YStep) {
		Rectangle bounds = poly1.getBounds();
		int minX = bounds.x;
		int minY = bounds.y;
		int width = bounds.width;
		int height = bounds.height;


		for (int x = minX; x < minX + width; x = x + XStep) {
			for (int y = minY; y < minY + height; y = y + YStep) {
				if (poly1.contains(x,y) && poly2.contains(x,y)) return true;
			}
		}

		return false;
	}


    /**
     * Rotates this polygon around it's center.
     *
     * @param degrees The number of degrees clockwise the polygon will be rotated
     * @see <code>WShape.rotatePoint (double x, double y, double degrees) </code>for how the
     * points are actually rotated.
     */
	public void rotate (double degrees) {
		super.rotate(degrees);

		double radians = Math.toRadians(degrees);

		for (int i=0; i<nPoints; i++) {

			//Rotate the current Point
			Point2D.Double currentPoint = WShape.rotatePoint(XCoordinatesOffset[i], YCoordinatesOffset[i], degrees);

			XCoordinatesOffset[i] = currentPoint.x;
			YCoordinatesOffset[i] = currentPoint.y;

		}

	}


    /**
     * Draws the filled representation of this WPoly to the given Graphics object.
     *
     * @param g The Graphics object that this WPoly will be drawn to.
     */
	public void drawFilled (Graphics g) {
		if (colourSet) g.setColor(colour);
		g.fillPolygon(toPolygon());
	}


    /**
     * Removes the filled representation of this WPoly to the given Graphics object.
     *
     * @param g The Graphics object that this WPoly will be removed from.
     * @param removeColour The colour that it will be removed to.
     */
     public void removeFilled (Graphics g, Color removeColour) {
		Color oldColour = g.getColor();
		g.setColor(removeColour);
		g.fillPolygon(toPolygon());
		g.setColor(oldColour);
	}



    /**
     * Draws the hollow representation of this WPoly to the given Graphics object.
     *
     * @param g The Graphics object that this WPoly will be drawn to.
     */
	public void drawHollow (Graphics g) {
		if (colourSet) g.setColor(colour);
		g.drawPolygon(toPolygon());
	}

    /**
     * Removes the hollow representation of this WPoly to the given Graphics object.
     *
     * @param g The Graphics object that this WPoly will be removed from.
     * @param removeColour The colour that it will be removed to.
     */
	public void removeHollow (Graphics g, Color removeColour) {
		Color oldColour = g.getColor();
		g.setColor(removeColour);
		g.drawPolygon(toPolygon());
		g.setColor(oldColour);
	}


    /**
     * Converts the data for the points that are relitive to the center point
     * to absolute coordinates, as used by the java.awt.Polygon class.
     *
     * @return The java.awt.Polygon object that is a representation of this WPoly.
     */
	public Polygon toPolygon () {


		int [] XCoordinates = new int [nPoints];
		int [] YCoordinates = new int [nPoints];

		for (int i = 0; i < nPoints; i++) {
			XCoordinates[i] = (int) (XCoordinatesOffset[i] + XCenter);
			YCoordinates[i] = (int) (YCoordinatesOffset[i] + YCenter);
		}

		Polygon toReturn =  new Polygon(XCoordinates, YCoordinates, nPoints);

		return toReturn;

	}

    /**
     * Returns the String representation of this WPoly object, that is with the
     * point coordinates represented relitive to the center point.
     *
     * @return The String representation of this WPoly object.
     */
	public String toString () {
		String toReturnX = "x:";
		String toReturnY = "y:";
		for (int i = 0; i < nPoints; i++) {
			toReturnX = toReturnX + " " + XCoordinatesOffset[i];
			toReturnY = toReturnY + " " + YCoordinatesOffset[i];
		}
		return "WPoly" + "\n" + toReturnX + "\n" + toReturnY+ "\n" + "Center: " + XCenter + "," + YCenter + "\n" + "Number of points: " + nPoints;
	}

    /**
     * Returns the String representation of this WPoly as a java.awt.Polygon object,
     * that is with the point coordinates represented as absolute coordinates.
     *
     * @return The String representation of the java.awt.Polygon representation
     * of this WPoly object.
     */
	public String toStringAsPolygon () {
		return toStringPolygon(toPolygon());
	}

    /**
     * Returns the String representation of the given java.awt.Polygon object.
     *
     * @param poly The polygon whose data will be returned represented as a
     * String.
     * @return The String representation of the given java.awt.Polygon.
     */
	public static String toStringPolygon (Polygon poly) {
		String toReturnX = "x:";
		String toReturnY = "y:";
		for (int i = 0; i < poly.npoints; i++) {
			toReturnX = toReturnX + " " + poly.xpoints[i];
			toReturnY = toReturnY + " " + poly.ypoints[i];
		}
		return "Polygon" + "\n" + toReturnX + "\n" + toReturnY+ "\n" + "Number of points: " + poly.npoints;
	}


}