//WShape.java
package com.tanksoftware.graphics;

import java.awt.Graphics;
import java.awt.Color;
import java.awt.geom.Point2D;

/**
 * Exactally the same as WBasicShape, but it has some more Colour detail, and
 * a rotation method.
 *
 * <p>Title: WShape</p>
 * <p>Description: Adds another layer of frame work from the WBasicShape class.
 * This includes storage of the colour of the model, and a constructor.</p>
 * <p>An example of this class in use can be found in the Tank Software program Tank Commander 2</p>
 * <p>Copyright: Copyright (c) 2002, free non-commercial use is permitted.</p>
 * <p>Company: Tank Software</p>
 * @author William Denniss
 * @version 1.0
 */

public abstract class WShape extends WBasicShape {


	protected Color colour = null;
	protected boolean colourSet;


    /**
     * Initialises the shape to it's starting points.
     *
     * @param XCenter The X coordinate of the center to this shape.
     * @param YCenter The Y coordinate of the center to this shape.
     */
	public WShape (double XCenter, double YCenter) {
		this.XCenter = XCenter;
		this.YCenter = YCenter;
	}



    /**
     * Rotates a given point around it's center by a given number of degrees.
     * <br><br>
     * <b>Example:</b> If you have a shape whose center is at position (1,1), and a point of
     * that shape at position (1,2), and you wish to rotate that point 90 degrees
     * in a clockwise direction you would call this method like so:
     * rotatePoint(0,1,90) where the first value is the horisontal difference between the
     * center point and the point to be rotated, in this case 0, the second is
     * the verticle difference between the center point and the point to be rotated
     * in this case 1, and the final value is the number of degrees the point
     * will be rotated around the center in a clockwise direction, in this case 90.
     * If this method is called with those perameters, it will return a Point2D.Double
     * With as it's X value, the new horistal difference (in the example, this would be 1)
     * and as it's Y value, the new verticle difference (in this example, this would be 0)
     *
     * @param x The horisontal distance of the point to be rotated from the center
     * of the shape.
     * @param y The verticle distance of the point to be rotated from the center
     * of the shape.
     * @param degrees The number of degrees clockwise the point will be rotated around the
     * shapes center.  For anti-clockwise, simply use a negative number.
     * @return The new verticle and horisontal coordinates of the point relitive
     * to the center of the shape.  The verticle distance is reperesented by the X,
     * and horisontal by the Y variables of the Point2D.Double class.
     */
	public static Point2D.Double rotatePoint (double x, double y, double degrees) {
    	//Rotate 1 point where the given points are offsets to the center point

		double radians = Math.toRadians(degrees);

		//Calculate the radias from this point to the center
		double hypotnues = Math.sqrt(Math.pow(x,2) + Math.pow(y,2));

		if (x == 0) x = 0.00000001;
		if (y == 0) y = 0.00000001;

		//calculate the angle of this radius
		double angle = Math.atan(x / y);

		//add 180 if it is negative to account for the absolute values calculated when squaring
		if (y < 0) angle = angle + Math.toRadians(180);

		//increase the angle by the desired amount
		angle = angle + radians;

		//recalculate the points
		x =  (hypotnues * Math.sin(angle));
		y = (hypotnues * Math.cos(angle));

		return new Point2D.Double (x, y);
	}

    /**
     * Sets the colour of this shape.
     *
     * @param colour The colour that this shape will be set to
     */
	public void setColour (Color colour) {
		this.colour = colour;
		colourSet = true;
	}

    /**
     * Removes the colouring of this shape.
     */
	public void removeColour () {
		colourSet = false;
	}


    //Already documented in WBasicShape
	public abstract void drawFilled(Graphics g);
	public abstract void removeFilled(Graphics g, Color colour);


}