//Ini.java
package com.tanksoftware.io;

import java.util.List;
import java.util.LinkedList;
import java.util.ArrayList;

import java.io.BufferedReader;
import java.io.IOException;

/**
 * An Ini parser that stores user variables easily in a human-readable text format.
 *
 * <p>Title: Ini</p>
 * <p>Description: This Ini parser is a hassle free way of loading settings and data from
 * preformatted ini files.  These Red Alert 2 formatted ini files are convient
 * as they be interpereted by humans in a text editor, and the ordering of the
 * sections and keys is entirly irrelevent (therefore making it easier to
 * change your config specifications).</p>
 *
 * <p>An example of this class in use can be found in the Tank Software program WillPang2.</p>
 * <p>Copyright: Copyright (c) 2002, free non-commercial use is permitted.</p>
 * <p>Company: Tank Software</p>
 *
 * Version History:<br>
 * 1.0 - 4 January 2002: Ini Created<br>
 * 1.1 - 13 March 2002: Small Changes<br>
 * 1.2 - 5 April 2002: Code reworked, now much neater (but performs essentaily the same service)<br>
 *
 *
 * @author William Denniss
 * @version 1.2 - 5 April 2002
 *
 */
public class Ini {


	private Section iniSectionData;
	private String filename;

    /**
     * Creates a new Ini, given the ini file filename from which the data will
     * be read.
     *
     * @param filename The name of the ini file to be loaded.
     * @param fileSystem The Tank Software File System that the ini will be
     * loaded through.  Note that this can be null if you are not using one.
     * @throws IOException If there is an error reading the ini file.
     */
	public Ini (String filename, TSFS fileSystem) throws IOException {
		this.filename = filename;

		iniSectionData = new Section ();

		List sections = Ini.readINIFile(filename, fileSystem);

		for (int i = 0; i < sections.size(); i++) {
			SubSection current = (SubSection) sections.get(i);;
			try {
				System.out.println(current.name);
				iniSectionData.addSubSection(current);
			} catch (NullPointerException e) {
			}
		}
	}

    /**
     * Reads the ini data from the specified file and returns it as a List of
     * all of the SubSections found in that file.
     *
     * @param filename The name of the ini file to be loaded.
     * @param fileSystem The Tank Software File System that the ini will be
     * loaded through.  Note that this can be null if you are not using one.
     * @return A list of all of the SubSections that were found in the given ini file.
     * @throws IOException If there is an error reading the ini file.
     */
	public static List readINIFile (String filename, TSFS fileSystem) throws IOException {

		List toReturn = new LinkedList();

		System.out.println("Reading INI file " + filename + "...");

	//	try {

			BufferedReader br = null;

			if (fileSystem != null)
				br = fileSystem.loadFile(filename);
			else
				br = TSFS.TSFSFile.getFileSpecificBufferedReader(filename);


			String line = "";
			char firstChar = '0';

			line = br.readLine();
			firstChar = firstNonSpaceChar(line);
			String currentComments = null;

			while (line != null) {

				if (firstChar == '[') {

					Section newSection = new Section (line.substring(1, line.indexOf(']')), new ArrayList(), line.substring(line.indexOf(']')+1, line.length()));
					newSection.preNameComment = currentComments;
					currentComments = null;

					System.out.print(newSection.name + " ");
					toReturn.add(newSection);
					line = br.readLine();
					firstChar = firstNonSpaceChar(line);
			//preNameComment

					 while (firstChar != '[' && line != null) {
						if (firstChar == '?' || isComChar(firstChar) || (line.indexOf('=') == -1 && line.indexOf('-') == -1)) {
							if (currentComments == null)
								currentComments = line;
							else
								currentComments +=  "\n" + line;

						} else {
							if (currentComments != null) {
								newSection.addSubSection(new LineComment(currentComments));
								currentComments = null;
							}

							int equPos = line.indexOf('=');
							String opcode = "";
							String operand = "";
							String followingComment = "";


						int comChar = firstComChar(line);

						if (equPos == -1)
							equPos = line.indexOf('-');

						opcode = line.substring(0,equPos);

						if (comChar > 1) {
							int startComment;
							if (line.indexOf(' ') > -1)
								startComment = Math.min(comChar, line.indexOf(' ',equPos));
							else
								startComment = comChar;

							operand = line.substring(equPos+1,startComment);
							followingComment = line.substring(startComment, line.length());
						} else {
							operand = line.substring(equPos+1,line.length() );
						}

						//System.out.println(opcode + "|" + operand + "|" + followingComment + "|");
						newSection.addSubSection(new Key(opcode, operand, followingComment));

						}

						line = br.readLine();
						firstChar = firstNonSpaceChar(line);

					}

				} else {
					toReturn.add(new LineComment(line));
					line = br.readLine();
				}

				firstChar = firstNonSpaceChar(line);

			}


	//	} catch (IOException e) {
			//end of the file, no problem
	//		e.printStackTrace();
	//		System.out.println(e.getLocalizedMessage());
	//	}

		return toReturn;
	}

    /**
     * Gets main Section of this ini file underwhich all Sections of the ini
     * reside (which intern store all of their SubSections).
     *
     * @return The main Section of this ini file.
     */
	public Section getIniSectionData () {
		return iniSectionData;
	}

    /**
     * Gets the filename that this ini file was read from.
     *
     * @return The filename that this ini file was read from.
     */
	public String getFilename () {
		return filename;
	}

    /**
     * Finds the first character in the given string that isn't a space nor a tab.
     *
     * @param line The Stirng to be parsed.
     * @return The first character in that string that isn't a space nor a tab.
     */
	private static char firstNonSpaceChar (String line) {

		if (line == null)
			return '?';

		for (int i = 0; i < line.length(); i++) {
			//System.out.println("++" + line.charAt(i) + "||" + i + "|");
			if (line.charAt(i) != ' ' && line.charAt(i) != '\t')
				return line.charAt(i);
		}

		return '?';
	}

    /**
     * Finds the location in the given string of the first character that is considered
     * the start of a comment.
     *
     * @param line The Stirng to be parsed.
     * @return The index of the first character in that string that is a comment character.
     */
	private static int firstComChar (String line) {

		if (line == null)
			return -1;

		for (int i = 0; i < line.length(); i++) {
			//System.out.println("++" + line.charAt(i) + "||" + i + "|");
			if (isComChar(line.charAt(i)))
				return i;
		}

		return -1;
	}

    /**
     * Determines whether or not a given character is classed as a comment character.
     *
     * @param toCheck The char to check to see if it is a comment char.
     * @return true if it is a comment char, false if it isn't.
     */
	private static boolean isComChar (char toCheck) {
		if (toCheck == '*' || toCheck == ';' || toCheck == '?') //=='/'
			return true;
		return false;
	}

    /**
     * The ini file format representation of the String (can be read back into an
     * Ini object.
     *
     * @return The String representing this Ini object.
     */
	public String toString () {
		return iniSectionData.toString();
	}





	/**
	 * A SubSection of the Ini file.  This is the basic data storage structure of
     * the Ini file that all other parts utilise.
	 */
	public static abstract class SubSection {
		public String name;
		public Section parent;
		public String afterComment = "";

		public String getSubSectionName () {
			return name;
		}

		public void addParent (Section parent) {
			this.parent = parent;
		}

		public abstract String toString ();
	}

	/**
	 * A comment which takes up the entire line (ie, starts with ";")
	 */
	public static class LineComment extends SubSection {
		//public String comment = "";
		public LineComment (String comment) {
			name = null;
			this.afterComment = comment;
		}

		public String toString () {
			return afterComment + "\n";
		}

	}

	/**
	 * A Key-Value pair (eg "width=10")
	 */
	public static class Key extends SubSection {
		//public String comment = "";
		public String value;

		public Key (String keyName, String value, String comment) {
			this.name = keyName;
			this.value = value;
			this.afterComment = comment;

		}


		public String toString () {
			return name + "=" + value + afterComment + "\n";
		}

	}


    /**
     * A form of SubSection which is actually a main heading, that stores several
     * other SubSections.
     */
	public static class Section extends SubSection{

		public List subSections = null;// = new ArrayList();

		public String preNameComment = null;

		public Section () {
			this.subSections = new ArrayList();
			this.name = null;
			this.afterComment = null;
		}

		public Section (String sectionName, List subSections, String comment) {
			this.subSections = subSections;
			this.name = sectionName;
			this.afterComment = comment;
		}

		public void addSubSection (SubSection toAdd) {
			subSections.add(toAdd);
			toAdd.addParent(this);
		}

		public void removeKey (String subSectionNameToRemove) throws SubSectionNotFoundException{
			SubSection toRemove = getSubSection(subSectionNameToRemove);
			if (toRemove == null)
				throw new SubSectionNotFoundException ();
			else
				subSections.remove(toRemove);
		}

		public void removeValue (String valueToRemoveTheKeyOf) throws SubSectionNotFoundException{


			SubSection toRemove = null;

			for (int i = 0; i < subSections.size(); i++) {

				try {
					Key current = (Key) subSections.get(i);

					if (current.value.equals(valueToRemoveTheKeyOf)) {
						toRemove = current;
					}
				} catch (ClassCastException e) {
				}
			}

			if (toRemove != null)
				subSections.remove(toRemove);
		}


		public SubSection getSubSection (String subSectionName) throws SubSectionNotFoundException{
			SubSection currentSubSection;
			for (int i = 0; i < subSections.size(); i++) {
				currentSubSection = (SubSection) subSections.get(i);
				if (currentSubSection.getSubSectionName() != null)
					if (currentSubSection.getSubSectionName().equalsIgnoreCase(subSectionName))
						return currentSubSection;

			}
			throw new SubSectionNotFoundException();
		}

		public void setSubSectionValue (String key, String value) {

			try {
				Key current = (Key) getSubSection(key);
				current.value = value;
			} catch (SubSectionNotFoundException e) {
				this.addSubSection(new Key(key, value, ""));
			}
		}

		public String getSubSectionKey(String value) throws SubSectionNotFoundException {

			for (int i = 0; i < subSections.size(); i++) {

				try {
					Key current = (Key) subSections.get(i);

					if (current.value.equals(value))
						return current.name;
				} catch (ClassCastException e) {
				}

			}
			throw new SubSectionNotFoundException ();


		}

		// Returns the value of a key that exists in this subsection.
		public String getSubSectionValue(String key) throws SubSectionNotFoundException{
			//System.out.println(((Key) windowOptions.getSubSection("width")).value);
			return ((Key) getSubSection(key)).value;
		}

		public int getSubSectionValueInt(String key, int defaultValue) {
			//System.out.println(((Key) windowOptions.getSubSection("width")).value);
			try {
				return Integer.parseInt(((Key) getSubSection(key)).value);
			} catch (Exception e) {
				return defaultValue;
			}
		}
		public double getSubSectionValueDouble(String key, double defaultValue) {
			//System.out.println(((Key) windowOptions.getSubSection("width")).value);
			try {
				return Double.parseDouble(((Key) getSubSection(key)).value);
			} catch (Exception e) {
				return defaultValue;
			}
		}

		public String toString() {


			String toReturn = "";
			if (this.preNameComment != null)
				toReturn += this.preNameComment + "\n";
			if (this.name != null) {
				toReturn += "[" + this.name + "]" + this.afterComment + "\n";
			}

			for (int i = 0; i < subSections.size(); i++) {
				//toReturn += "SubSection: " + ((SubSection) subSections.get(i)).name;
				toReturn += ((SubSection) subSections.get(i)).toString();
			}
			return toReturn;
		}
	}




    /**
     * Thrown when a SubSection is searched for but doesn't exist.
     */
	public static class SubSectionNotFoundException extends Exception {
	}

    /**
     * Thrown when a Section is searched for but doesn't exist.
     */
	public static class SectionNotFoundException extends Exception {
	}
}